/*************************************************************************
 cpp file for IoInterface class;
 Autor :
         Yu Lu
         luyu1974@gmail.com
 *************************************************************************/

#include "Gps_interface.h"
//---------------------------------------------------------------------------
/*++    Routine Description:
    Constructor of class IoInterface;
    Duty:
        Open the handle, if success, set handleOpened true;
-- */

IoInterface::IoInterface()
{
 count = 0;
 ping_pong_flag = 0;
 memset( (void *)gp2021regs, 0, 2*sizeof(gp2021_regs));
 infofile.open(".\\log\\gps_interface.log",ios::out);
 gpsHandle = OpenHandle();
 if( gpsHandle != INVALID_HANDLE_VALUE )
 {
    handleOpened = true;
 }
 else
 {
   s_thread = NULL;
   handleOpened = false;
 }
}

//---------------------------------------------------------------------------
/*++
    Routine Description:
    Destructor of class interface;
    if the gpsHandle is opened, close it;
--*/

 IoInterface::~IoInterface()
{
 if(handleOpened)
  CloseHandle(gpsHandle);

 if(s_thread)
 {
  delete s_thread;
 }
}
//---------------------------------------------------------------------------
/*++
   Routine Description:
   Init proc for class;
   Return non-zero if something wrong, otherwise return 0
   Error code is defined in gbl_var.h
 --*/

int IoInterface::InitProc()
{
  if(!infofile)
  return(IOINTERFACE_FILEOPENERR);
  return (IOINTERFACE_SUCCESS);
}

//---------------------------------------------------------------------------
/*++
  Routine Description:
  print debug info for class;
--*/

void IoInterface::PrintDbgInfo(char* info)
{
   if(infofile)
    infofile << info << endl;
}

//---------------------------------------------------------------------------
/*++
  Routine Description:
  return the sataus of handleOpened;
--*/


bool IoInterface::IsHandleOpened()
{
 return handleOpened;
}
//---------------------------------------------------------------------------
HANDLE  IoInterface::OpenOneDevice(HDEVINFO HardwareDeviceInfo,
                                  PSP_INTERFACE_DEVICE_DATA DeviceInfoData,
                                  char *devName)
/*++
Routine Description:

    Given the HardwareDeviceInfo, representing a handle to the plug and
    play information, and deviceInfoData, representing a specific usb device,
    open that device and fill in all the relevant information in the given
    USB_DEVICE_DESCRIPTOR structure.

Arguments:

    HardwareDeviceInfo:  handle to info obtained from Pnp mgr via SetupDiGetClassDevs()
    DeviceInfoData:      ptr to info obtained via SetupDiEnumInterfaceDevice()

Return Value:

    return HANDLE if the open and initialization was successfull,
	else INVLAID_HANDLE_VALUE.

--*/
/***************************************************************************/
{
 char buf[512];
    PSP_INTERFACE_DEVICE_DETAIL_DATA     functionClassDeviceData = NULL;
    ULONG                                predictedLength = 0;
    ULONG                                requiredLength = 0;
    HANDLE  hOut = INVALID_HANDLE_VALUE;

    //
    // allocate a function class device data structure to receive the
    // goods about this particular device.
    //
    SetupDiGetInterfaceDeviceDetail(
            HardwareDeviceInfo,
            DeviceInfoData,
            NULL, // probing so no output buffer yet
            0, // probing so output buffer length of zero
            &requiredLength,
            NULL); // not interested in the specific dev-node


    predictedLength = requiredLength;
    // sizeof (SP_FNCLASS_DEVICE_DATA) + 512;

    functionClassDeviceData = (PSP_INTERFACE_DEVICE_DETAIL_DATA)malloc(predictedLength);
    functionClassDeviceData->cbSize = sizeof(SP_INTERFACE_DEVICE_DETAIL_DATA);

    //
    // Retrieve the information from Plug and Play.
    //
    if (! SetupDiGetInterfaceDeviceDetail(
               HardwareDeviceInfo,
               DeviceInfoData,
               functionClassDeviceData,
               predictedLength,
               &requiredLength,
               NULL)) {
		free( functionClassDeviceData );
        return INVALID_HANDLE_VALUE;
    }
    strcpy( devName,functionClassDeviceData->DevicePath) ;

    sprintf(buf, "Attempting to open: %s", devName );
    PrintDbgInfo(buf);

    hOut = CreateFile(
                  functionClassDeviceData->DevicePath,
                  GENERIC_READ | GENERIC_WRITE,
                  FILE_SHARE_READ | FILE_SHARE_WRITE,
                  NULL, // no SECURITY_ATTRIBUTES structure
                  OPEN_EXISTING, // No special create flags
                  0, // No special attributes
                  NULL); // No template file

    if (INVALID_HANDLE_VALUE == hOut) {

                sprintf(buf, "FAILED to CreateFile %s", devName);
                PrintDbgInfo(buf);
    }
	free( functionClassDeviceData );
	return hOut;
}

//---------------------------------------------------------------------------
HANDLE  IoInterface::OpenUsbDevice( LPGUID  pGuid, char *outNameBuf)
/*++
Routine Description:

   Do the required PnP things in order to find
   the next available proper device in the system at this time.

Arguments:

    pGuid:      ptr to GUID registered by the driver itself
    outNameBuf: the generated name for this device

Return Value:

    return HANDLE if the open and initialization was successful,
	else INVLAID_HANDLE_VALUE.
--*/
/***************************************************************************/
{
   ULONG NumberDevices;
   HANDLE hOut = INVALID_HANDLE_VALUE;
   HDEVINFO                 hardwareDeviceInfo;
   SP_INTERFACE_DEVICE_DATA deviceInfoData;
   ULONG                    i;
   BOOLEAN                  done;
   PUSB_DEVICE_DESCRIPTOR   usbDeviceInst;
   PUSB_DEVICE_DESCRIPTOR	*UsbDevices = &usbDeviceInst;
   char buf[512];

   *UsbDevices = NULL;
   NumberDevices = 0;

   //
   // Open a handle to the plug and play dev node.
   // SetupDiGetClassDevs() returns a device information set that contains info on all
   // installed devices of a specified class.
   //
   hardwareDeviceInfo = SetupDiGetClassDevs(
                           pGuid,
                           NULL, // Define no enumerator (global)
                           NULL, // Define no
                          (DIGCF_PRESENT | // Only Devices present
                           DIGCF_INTERFACEDEVICE)); // Function class devices.

   //
   // Take a wild guess at the number of devices we have;
   // Be prepared to realloc and retry if there are more than we guessed
   //
   NumberDevices = 4;
   done = false;
   deviceInfoData.cbSize = sizeof(SP_INTERFACE_DEVICE_DATA);

   i=0;
   while (!done) {
      NumberDevices *= 2;

      if (*UsbDevices) {
         *UsbDevices =
               (PUSB_DEVICE_DESCRIPTOR)realloc(*UsbDevices, (NumberDevices * sizeof(USB_DEVICE_DESCRIPTOR)));
      } else {
         *UsbDevices = (PUSB_DEVICE_DESCRIPTOR)calloc(NumberDevices, sizeof(USB_DEVICE_DESCRIPTOR));
      }

      if (NULL == *UsbDevices) {

         // SetupDiDestroyDeviceInfoList destroys a device information set
         // and frees all associated memory.

         SetupDiDestroyDeviceInfoList(hardwareDeviceInfo);
         return INVALID_HANDLE_VALUE;
      }

      usbDeviceInst = *UsbDevices + i;

      for (; i < NumberDevices; i++) {

         // SetupDiEnumDeviceInterfaces() returns information about device interfaces
         // exposed by one or more devices. Each call returns information about one interface;
         // the routine can be called repeatedly to get information about several interfaces
         // exposed by one or more devices.

         if( SetupDiEnumDeviceInterfaces(hardwareDeviceInfo,
                                         0,
					 pGuid,
                                         i,
                                         &deviceInfoData))
           {
            hOut = OpenOneDevice(hardwareDeviceInfo, &deviceInfoData, outNameBuf);
	    if ( hOut != INVALID_HANDLE_VALUE ) {
               done = TRUE;
               break;
	    }
           } else {      // 'else' of "if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,"
             sprintf(buf, "Error of SetupDiEnumDeviceInterfaces");
             PrintDbgInfo(buf);

             if(ERROR_NO_MORE_ITEMS == GetLastError())
              {
               done = TRUE;
               break;
              }
         }  // end of "if (SetupDiEnumDeviceInterfaces (hardwareDeviceInfo,"
      }   // end of "for (; i < NumberDevices; i++) {"
   }  // end of "while (!done) {"

   NumberDevices = i;

   // SetupDiDestroyDeviceInfoList() destroys a device information set
   // and frees all associated memory.

   SetupDiDestroyDeviceInfoList(hardwareDeviceInfo);
   free ( *UsbDevices );
   return hOut;
}

//---------------------------------------------------------------------------
/***************************************************************************/
HANDLE IoInterface::OpenHandle( void )
{
 char buf[512];
 HANDLE hDEV = OpenUsbDevice( (LPGUID)&GUID_INTERFACE_USBGPS, completeDeviceName);
	if (hDEV == INVALID_HANDLE_VALUE) {
               sprintf(buf,"Failed to open %s %s", completeDeviceName, GetLastError());
               PrintDbgInfo(buf);
               /*
                ifofile << "\n Failed to open " << completeDeviceName << GetLastError();
               */
        } else {
                sprintf(buf,"Succeeding to open: %s %s", completeDeviceName, GetLastError());
                PrintDbgInfo(buf);
               }
        return hDEV;
}

//---------------------------------------------------------------------------
/*++
  Routine Description:
  clear gp2021regs to all zeros;
--*/


void IoInterface::clear_all_reg(unsigned char bufferflag)
{
 assert( bufferflag ==0 || bufferflag == 1);

 memset( (void *)(gp2021regs+ bufferflag), 0, sizeof(gp2021_regs));
}

//---------------------------------------------------------------------------
/*++
  Routine Description:
  To read the register of USBN9603 , just for debugging
  Argument:
          addr:  the address of register to be read;
          out :  the pointer that store the value of the register
  return:
          if successful , rutren true; otherwise return false
--*/

bool IoInterface::read_usbn9603_reg(const unsigned char addr, unsigned char *out)
{
 char infobuf[256];
 unsigned char buf[6]={0xaa, 'g', 0,0,0,0};
 int nBytes;
  if( handleOpened )
   {
	buf[4] = addr;
	buf[5] = (buf[0]+buf[1]+buf[4]) % 256;  // this is the checksum

	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_WRITE,
			buf,
			6,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured writing data to usb handle");
                PrintDbgInfo(infobuf);
		return false;
	}

	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_READ,
			NULL,
			0,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured reading data from usb handle ");
                PrintDbgInfo(infobuf);
	return false;
	}

	*out = buf[0];
	return true;
  } // end of if( fileOpened )
  else
  return false;
}

//---------------------------------------------------------------------------
/*++
  Routine Description:
  To read all the registers of USBN9603 , just for debugging
  Argument:
          out :  the pointer that store the values of the registers
  return:
          if successful , rutren true; otherwise return false
--*/

bool IoInterface::read_usbn9603_allreg( unsigned char *out)
{
 char infobuf[256];
 unsigned char buf[6]={0xaa, 'b', 0,0,0,0}, reg_buf[64];
 int nBytes;
  if( handleOpened )
   {
	buf[5] = (buf[0]+buf[1]) % 256;  // this is the checksum

	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_WRITE,
			buf,
			6,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured writing data to usb handle ");
                PrintDbgInfo(infobuf);
		return false;
	}

	if(!DeviceIoControl(gpsHandle,
 			IOCTL_USBGPS_BULK_READ,
			NULL,
			0,
			reg_buf,
			63,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured reading data from usb handle ");
                PrintDbgInfo(infobuf);
	 return false;
	}
        memcpy(out,reg_buf,63);
	return true;
  } // end of if( fileOpened )
  else
  return false;
}


//---------------------------------------------------------------------------
/*++
  Routine Description:
  To read the register of GP2021
  Argument:
          addr:  the address of register to be read;
          out :  the pointer that store the value of the register (16bit)
  return:
          if successful , rutren true; otherwise return false
--*/

bool IoInterface::read_gp2021_reg(const unsigned char addr, unsigned int *out)
{
 char infobuf[256];
 unsigned char buf[6]={0xaa, 'q', 0,0,0,0};
 int nBytes;
 if( handleOpened )
 {
	buf[2] = addr;
	buf[5] = (buf[0]+buf[1]+buf[2]) % 256; // calculate the checksum

	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_WRITE,
			buf,
			6,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured writing data to usb handle ");
                PrintDbgInfo(infobuf);
		return false;
	}


   	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_READ,
			NULL,
			0,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured reading data from usb handle ");
                PrintDbgInfo(infobuf);
	return false;
	}
	*out = buf[1]*256 + buf[0];
	return true;
  } // end of if( handleOpened )
  else
  return false;
}
//---------------------------------------------------------------------------
/*++
  Routine Description:
  To write to GP2021 regs ,
  Argument:
          addr :  address of GP2021 reg
          in   :  value to be written
  return:
          if successful , rutren true; otherwise return false
--*/
bool IoInterface::write_gp2021_reg(const unsigned char addr, const unsigned int in)
{
  char infobuf[256];
  unsigned char buf[6]={0xaa, 't', 0,0,0,0};
  int nBytes;
  if(handleOpened)
  {
	buf[2] = addr;
	buf[3] = in & 0xff;
	buf[4] = (in >> 8) & 0xff;
	buf[5] = (buf[0]+buf[1]+buf[2]+buf[3]+buf[4]) % 256;   // checksum

	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_WRITE,
			buf,
			6,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured writing data to usb handle ");
                PrintDbgInfo(infobuf);
		return false;
	}

	return true;
   } // end of if(handleOpened)
   else
   return false;
}

//---------------------------------------------------------------------------
/*++
  Routine Description:
  To read all the registers of GP2021 , just for debugging
  Argument:
          out :  the pointer that store the values of the registers
  return:
          if successful , rutren true; otherwise return false
--*/

bool IoInterface::read_gp2021_all( unsigned int *out)
{
  char infobuf[256];
  unsigned char buf[6]={0xaa, 'a', 0,0,0,0};
  unsigned char outbuf[449];
  int nBytes;
     if( handleOpened )
     {
        buf[5] = (buf[0]+buf[1]) % 256;

	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_WRITE,
			buf,
			6,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured get_all_gp2021 to usb handle ");
                PrintDbgInfo(infobuf);
		return false;
	}


   	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK1_READ,       // From TXD3 , change this also requires firmware changes
                        //IOCTL_USBGPS_BULK_READ, // From TXD1
			NULL,
			0,
			outbuf,
			449,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured reading get_all_gp2021 data from usb handle, %d ", nBytes);
                PrintDbgInfo(infobuf);
		return false;
	}
        for( int i = 0 ; i<188; i++)
        *(out + i) = outbuf[2*i] + outbuf[2*i + 1] * 256;
	return true;
      }  // end of ( if( handleOpened ))
      else
      return false;
}

//---------------------------------------------------------------------------
/*++
  Routine Description:
  To write a bunch of data to GP2021 , just for debugging
  Argument:
          in :  the pointer that store the values to be written
          size: size of data buffer
  return:
          if successful , rutren true; otherwise return false
--*/
bool IoInterface::write_gp2021_all( unsigned char *in, unsigned int size)
{
  int nBytes;
  char infobuf[256];
       if( handleOpened )
       {
    	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK1_WRITE,
			in,
			size,
			NULL,
			0,
			(unsigned long *)&nBytes,
			NULL) )
	{
                sprintf(infobuf, "An error occured writing all data to usb handle ");
                PrintDbgInfo(infobuf);
		return false;
	}

	return true;
       } // end of  if( handleOpened )
       else
       return false;
}
//---------------------------------------------------------------------------
/*++
  Routine Description:
  To send command to GP2021 , then try to get the 4 collections;

  Argument:
          bufferfalg :  indicate which data buffer is transferring;
          needreturn :  if app need the 4 collections;
          out        : point to store the 4 collection if needreturn == true;
  return:
          if successful , rutren true; otherwise return false
--*/

bool IoInterface::rd_wr_gp2021_all(char bufferflag,bool needreturn, unsigned char *out)
{
 int nBytes;
 char infobuf[256];
 // every reg needs 3 bytes, plus this count byte

 assert( bufferflag == 0 || bufferflag ==1);

  if( handleOpened )
  {
   if( needreturn )
   {
    gp2021regs[bufferflag].flag = gp2021regs[bufferflag].flag | 0x01;  // needs to get return
    if(!DeviceIoControl(gpsHandle,
          		IOCTL_USBGPS_BULK1_RD_WR,
			(unsigned char*)(gp2021regs+bufferflag),
			sizeof(gp2021_regs),
			out,
			LENG_OF_ONE_FRAME,
			(unsigned long *)&nBytes,
			NULL) )
	{
         sprintf(infobuf, "An error occured writing & reading all data to usb handle: %d count: %d", nBytes, count);
         PrintDbgInfo(infobuf);

	return false;
	}
    }    // match of  if( needreturn)
    else   // don't need return , just send command to gp2021
    {
     gp2021regs[bufferflag].flag = gp2021regs[bufferflag].flag & 0xfe;
     	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK1_WRITE,
			(unsigned char*)(gp2021regs+bufferflag),
			sizeof(gp2021_regs),
			NULL,
			0,
			(unsigned long *)&nBytes,
			NULL) )
	{
         sprintf(infobuf, "An error occured writing all data to usb handle ");
         PrintDbgInfo(infobuf);

	 return false;
	}
    }  // end of if(needreturn)

    return true;
   } // match of  if( handleOpened )
   else
   return false;
}
//---------------------------------------------------------------------------
/*++
  Routine Description:
  control the timer of DS89c420,
  Argument:
      flag = 0:  reset and stop timer,
                thus stop collecting gp2021's tracking loop data
      flag = 1:  restart timer,
                thus start collecting gp2021's tracking loop data continuously
  return:
          if successful , rutren true; otherwise return false
--*/
bool IoInterface::control_gp2021(char flag)
{

 char infobuf[256];
 unsigned char buf[6]={0xaa, 'c', 0,0,0,0};
 int nBytes;
 if( handleOpened )
 {
	buf[2] = flag;
	buf[5] = (buf[0]+buf[1]+buf[2]) % 256; // calculate the checksum

	if(!DeviceIoControl(gpsHandle,
			IOCTL_USBGPS_BULK_WRITE,
			buf,
			6,
			buf,
			6,
			(unsigned long *)&nBytes,
			NULL) )
	{
          sprintf(infobuf, "An error occured writing data to usb handle ");
          PrintDbgInfo(infobuf);
	  return false;
	}
        return true;
 }
 else
 return false;
}
//---------------------------------------------------------------------------
/*++
  Routine Description:
  Wrapping of rd_wr_gp2021_all for application,
  Argument:
      pp_flag :  indicate the ping_pong flag
  return:
          if successful , rutren true; otherwise return false
--*/
bool IoInterface::WriteReadGP2021(unsigned char pp_flag)
{
// unsigned char outbuf[LENG_OF_ONE_FRAME];
 bool success;
 count ++;
 success = rd_wr_gp2021_all(pp_flag,true,tracking_data[pp_flag]);
 return( success );
}
//---------------------------------------------------------------------------
/*++
  Routine Description:
  copy collection to GPS_Nav thread, and switch pp_flag
  Argument:
         des: pointer of command buffer
  return:
         none;
--*/

void IoInterface::CopyData(unsigned char* des)
{
 // ping_pong_flag is current working buffer, so we need to toggel it
 unsigned char flag = (ping_pong_flag+1)%2;
 assert(des != NULL);
 memcpy((void*)des, (void*)tracking_data[flag], LENG_OF_ONE_FRAME);
}
//---------------------------------------------------------------------------
/*++
  Routine Description:
  copy command from GPS_Nav thread to current idle buffer,
  Argument:
      bufferflag :  indicate which buffer should receive this data buffer,

      src     :  pointer of data buffer
  return:
          none;
--*/
void IoInterface::GetCntlData(char bufferflag, unsigned char* src)
{

 assert(bufferflag == 0 || bufferflag == 1);
 assert(src != NULL);
 memcpy((void*)(gp2021regs+bufferflag), (void*)src, sizeof(gp2021_regs));
}

//---------------------------------------------------------------------------
/* function called in running thread ,                                     */
/*  its duty should set ping-pong flag to indicate available buffer        */
/***************************************************************************/

void IoInterface::set_pp_flag( char PPflag)
{
 ping_pong_flag = PPflag;
}
unsigned char IoInterface::toggle_pp_flag(void)
{
 ping_pong_flag  = (ping_pong_flag+1)%2;
 return ping_pong_flag;
}
unsigned char IoInterface::get_pp_flag(void)
{
 return ping_pong_flag;
}
unsigned char IoInterface::get_free_pp_flag(void)
{
 // ping_pong_flag indicate the buffer which is in use currently
 return (ping_pong_flag+1)%2;
}

//---------------------------------------------------------------------------
/*    Routine Description:                                                 */
/*    thread handling funtcions                                          */
/***************************************************************************/

void IoInterface::CreatSampleThrd()
{
 if(handleOpened)
 {
  clear_all_reg(0);
  clear_all_reg(1);
  s_thread = new SamplingThread(true, this);
  s_thread->Resume();
  thrdtermed = false;
 }
}
// terminate sampling thread
void IoInterface::ThrdTerminate()
{
 if(s_thread)
  s_thread->Terminate();
}
// return the terminate status of sampling thread
bool IoInterface::NavThrdIsTermed()
{
 return (thrdtermed);
}
 // function to set the terminate status of sampling thread
void IoInterface::SetThrdStatus(bool stat)
{
 thrdtermed = stat;
}
