

#include "Gpsnav.h"
/*******************************************************************************
FUNCTION exor(char bit, long parity)
RETURNS  None.

PARAMETERS
			bit     char
			parity  long

PURPOSE
			count the number of bits set in the parameter parity and
			do an exclusive or with the parameter bit

WRITTEN BY
	Clifford Kelley
CHANGED BY
        Yu Lu

*******************************************************************************/
static int exor(char bit, unsigned int parity)
{
  char i;
  int result, tmp_par;
  result=0;
  tmp_par = parity>>6;
  for (i=0;i<24;i++)
  {
    if( tmp_par & 0x1) result++;
    tmp_par = tmp_par>>1;
  }
  result=result%2;

  result=(bit ^ result) & 0x1;
  return(result);
}
/*******************************************************************************
FUNCTION xors(long pattern)
RETURNS  Integer

PARAMETERS
			  pattern  long 32 bit data

PURPOSE
			  count the number of bits set in "pattern"
WRITTEN BY
	Clifford Kelley
CHANGED  by
         Yu Lu

*******************************************************************************/
static int xors(long pattern)
{
  int count,i;
  count=0;
  pattern=pattern>>6;
  for (i=0;i<=25;i++)
  {
	 count+=int(pattern & 0x1);
	 pattern=pattern>>1;
  }
  count=count%2;
  return(count);
}

static int sign(double a)
{
 if( a>0) return 1;
 else return -1;
}

static int bit_sign(double a)
{
 if( a>0 ) return 1;
 else return 0;
}
//---------------------------------------------------------------------------
/*    Routine Description:                                                 */
/*  Constructor of class GPS_Nav_class;                                    */
/***************************************************************************/

GPS_Nav::GPS_Nav(IoInterface* Iohandle)
{
 IoHandle = Iohandle;
 navthrd = NULL;
 newfrmflag = false;
 newfrmcount = 0;
 thrdtermed= true;
 memset((void*)&frm_4ms_6ch, 0, sizeof(one_frame_data));
 count = 0;
 count1 = 0;
 TIC_count =  0;
 ch_status = 0;
 for(char i=0; i<CH_NUM ; i++)
  ch[i].subframe_i = -1;

}

/*  Destructor of class GPS_Nav_class;  */  
GPS_Nav::~GPS_Nav()
{
 if(navthrd) {
  delete navthrd;
 }

#ifdef GPS_NAV_DBG
 debugfile.close();
#endif

#if GPS_NAV_RAW_DATA > 0
 ch_accum_mis_file.close();
 for(int i=0; i<CH_NUM; i++)
  ch_rawdata_file[i].close();
#endif
}
//---------------------------------------------------------------------------
/* Routine Description:
  To add one reg to be written to the gp2021regs
  Argument:
          add:  the address of register to be written;
          hi :  high byte
          lo : low byte
  return:
          no
*/
/***************************************************************************/
void GPS_Nav::add_one_reg( unsigned char add,
                           unsigned char hi,
                           unsigned char lo)
{
 unsigned char reg_count;

 if(gp2021_wr_data.count < 41)
 {
  reg_count = gp2021_wr_data.count++;
  gp2021_wr_data.regs[reg_count].addr = add;
  gp2021_wr_data.regs[reg_count].low  = lo;
  gp2021_wr_data.regs[reg_count].high = hi;
 }
}
void GPS_Nav::clear_all_reg()
{
 memset( (void *)(&gp2021_wr_data), 0, sizeof(gp2021_regs));
}
/***************************************************************************/
/* GP2021  related function *********/
void GPS_Nav::ch_cntl(char ch,  unsigned short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = (ch+CH_OFFSET)<<3;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}

void GPS_Nav::ch_code( char ch,unsigned int freq)
{
 unsigned short high_freq, low_freq;
 high_freq = (freq>>16) & 0xffff;
 low_freq  =  freq & 0xffff;

 // first is high 16-bit freq word
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = ((ch+CH_OFFSET)<<3)+5;
 gp2021_wr_data.regs[idx].high = (high_freq>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  high_freq & 0xff;

 // next is low 16-bit freq word
 assert(gp2021_wr_data.count < 41);
 idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = ((ch+CH_OFFSET)<<3)+6;
 gp2021_wr_data.regs[idx].high = (low_freq>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  low_freq & 0xff;
}

void GPS_Nav::ch_carrier(char ch,unsigned  int freq)
{
 unsigned short high_freq, low_freq;
 high_freq = (freq>>16) & 0xffff;
 low_freq  =  freq & 0xffff;

 // first is high 16-bit freq word
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = ((ch+CH_OFFSET)<<3)+3;
 gp2021_wr_data.regs[idx].high = (high_freq>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  high_freq & 0xff;

 // next is low 16-bit freq word
 assert(gp2021_wr_data.count < 41);
 idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = ((ch+CH_OFFSET)<<3)+4;
 gp2021_wr_data.regs[idx].high = (low_freq>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  low_freq & 0xff;
}
void GPS_Nav::ch_epoch_load(char ch, unsigned short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = ((ch+CH_OFFSET)<<3)+7;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}
void GPS_Nav::ch_on(char ch)
{
 ch_status=ch_status | bit_pat[ch+CH_OFFSET];
}
void GPS_Nav::ch_off(char ch)
{
 ch_status=ch_status & !bit_pat[ch+CH_OFFSET];
}
void GPS_Nav::set_ch_onoff(void)
{
 reset_cntl(ch_status);
}
void GPS_Nav::test_control(unsigned  short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = 0x7c;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}
void GPS_Nav::system_setup(unsigned  short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = 0x7e;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}
void GPS_Nav::io_config(unsigned  short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = 0xf0;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}
void GPS_Nav::reset_cntl(unsigned  short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = 0x7f;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}
void GPS_Nav::program_TIC(unsigned  int data)
{
 unsigned short high_16, low_16;
 high_16 = (data>>16) & 0xffff;
 low_16  =  data & 0xffff;

 // first is high 16-bit  word
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = 0x6d;
 gp2021_wr_data.regs[idx].high = (high_16 >> 8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  high_16 & 0xff;

 // next is low 16-bit  word
 assert(gp2021_wr_data.count < 41);
 idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = 0x6f;
 gp2021_wr_data.regs[idx].high = (low_16>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  = (low_16) & 0xff;
}
void GPS_Nav::ch_accum_reset(char ch)
{
}
void GPS_Nav::ch_code_slew(char ch, unsigned short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = ((ch+CH_OFFSET)<<2)+0x84;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}

void GPS_Nav::test_data(unsigned  short data)
{
 assert(gp2021_wr_data.count < 41);
 char idx = gp2021_wr_data.count++;
 gp2021_wr_data.regs[idx].addr = 0xf2;
 gp2021_wr_data.regs[idx].high = (data>>8) & 0xff;
 gp2021_wr_data.regs[idx].low  =  data & 0xff;
}

void GPS_Nav::get_newaccumdata(void)
{
 double q_sum, i_sum, tmp_theta;

 epoch_chk  = 1;
 for( int j=0; j<4; j++)
  if((frm_4ms_6ch.all_ch[j].accum_status_b) & 0x2000 ) // TIC happened here
   epoch_chk = 0;      // epoch_check values are not available

 for( int i=0; i< CH_NUM; i++)
  ch[i].valid_data_num = 0;

 for(int i=0; i<CH_NUM; i++)
 {
  for(int j=0; j<4; j++)
  {
   if( (frm_4ms_6ch.all_ch[j].accum_status_a) & (ch_ready_mask[i+CH_OFFSET]) )
   {
    // we have valid data from channel i in jth packet;
    ch[i].prompt_iq[ ch[i].valid_data_num++ ][0] = frm_4ms_6ch.all_ch[j].ch_accm[i].I_prompt;
    ch[i].prompt_iq[ ch[i].valid_data_num -1 ][1] = frm_4ms_6ch.all_ch[j].ch_accm[i].Q_prompt;
    ch[i].track_iq[ ch[i].valid_data_num -1 ][0] = frm_4ms_6ch.all_ch[j].ch_accm[i].I_track;
    ch[i].track_iq[ ch[i].valid_data_num -1 ][1] = frm_4ms_6ch.all_ch[j].ch_accm[i].Q_track;
    if( epoch_chk )
     ch[i].epochchk[ ch[i].valid_data_num -1 ] = *((unsigned short*)(&frm_4ms_6ch.ch_cntl[0])+j*6+ i);

   }   // end of if( (frm_4ms_6ch.all_ch[j
  }    // end of j  for(int j=0; j<4; j++){
  ch[i].mean_p=0.;
  ch[i].mean_t=0.;
  for( int k=0; k< ch[i].valid_data_num; k++)
  {
   ch[i].track_mag[k] = sqrt(pow((float)(ch[i].track_iq[ k ][0]),2)+pow((float)(ch[i].track_iq[ k ][1]),2));
   ch[i].prompt_mag[k]= sqrt(pow((float)(ch[i].prompt_iq[ k ][0]),2)+pow((float)(ch[i].prompt_iq[ k ][1]),2));

   ch[i].mean_p += ch[i].prompt_mag[k];
   ch[i].mean_t += ch[i].track_mag[k];
   q_sum = ch[i].prompt_iq[k][1];// + ch[i].track_iq[k][1];
   i_sum = ch[i].prompt_iq[k][0];// + ch[i].track_iq[k][0];

   if( ch[i].state >= PULLIN)
   {
    if( q_sum != 0. || i_sum != 0.)
     ch[i].theta[k] = atan2( q_sum, i_sum);
    else
    {
     if( k == 0)
      ch[i].theta[k] = ch[i].old_theta;
     else
      ch[i].theta[k] = ch[i].theta[k-1];
    }
   }  //  end of if( ch[i].state

   if( ch[i].state == TRACKING)
   {
    if( i_sum !=0 || q_sum != 0)
     ch[i].ca_phase[k] = ch[i].theta[k]-(PI/2)*sign(q_sum);
//     ch[i].ca_phase[k] = -i_sum*sign(q_sum)/sqrt(i_sum*i_sum + q_sum*q_sum);
   }
  }   // end of k
  ch[i].mean_p = ch[i].mean_p/ch[i].valid_data_num;
  ch[i].mean_t = ch[i].mean_t/ch[i].valid_data_num;

  if( ch[i].state >= PULLIN)
  {
   ch[i].cd_phase1 =
   (ch[i].mean_p - ch[i].mean_t)/(ch[i].mean_p + ch[i].mean_t)*pull_code_k;
  }
 }  // end of for(int i=0;
}
void GPS_Nav::process_channel( int idx )
{
 assert( idx < CH_NUM && idx >=0);

 switch( ch[idx].state){
  case ACQUISITION: ch_acq(idx); break;
  case CONFIRM: ch_confirm(idx); break;
  case PULLIN: ch_pullin(idx); break;
  case TRACKING: ch_track(idx); break;
  default: break;
 }
}

void  GPS_Nav::ch_acq(int idx)
{
 assert( idx < CH_NUM && idx >=0);
 if( ch[idx].mean_p < CONFIRM_THRESHOLD || ch[idx].mean_t < CONFIRM_THRESHOLD )
 {
  if(ch[idx].code > 2044)
  {
   ch[idx].code =0;
   ch[idx].n_freq ++;
   ch_carrier(idx,carrier_ref+(int)(4698*ch[idx].n_freq));
   ch[idx].carrier_freq = carrier_ref+(int)(4698*ch[idx].n_freq);

   if(ch[idx].n_freq >= SEARCH_MAX_F)
   {
    ch[idx].n_freq = SEARCH_MIN_F;
   }
  }
  else
  {
   ch_code_slew(idx,2);
   ch[idx].code += 2;
  }
 }
 else
 {
  ch_code_slew(idx,2042);  // slide back 2 chips , because of frm delay
  ch[idx].state = CONFIRM;
  ch[idx].n_confirm = 0;
  ch[idx].mn_confirm = 0;
 }
}
void  GPS_Nav::ch_confirm(int idx)
{
 assert( idx < CH_NUM && idx >=0);

 ch[idx].n_confirm++;
 if( ch[idx].mean_p > CONFIRM_THRESHOLD)// && ch[idx].mean_t > 500. )
  ch[idx].mn_confirm++;

 if(ch[idx].n_confirm == 7)
 {
  if( ch[idx].mn_confirm < 4 )
  {
   ch[idx].state = ACQUISITION;
   ch[idx].mn_confirm =0;
   ch[idx].n_confirm =0;
  }
  else
  {
   ch[idx].state = PULLIN;
   ch[idx].mn_confirm =0;
   ch[idx].n_confirm =0;
   ch[idx].cd_phase0 = ch[idx].cd_phase1; // prepare for PULLIN

   ch[idx].sum_prompt_mag = 0.;
   ch[idx].sum_delta_theta = 0.;
   ch[idx].pullin_time_idx = 0;
   ch[idx].pull_time_count =0;
   ch[idx].ms_count = 0;
   ch[idx].ms_set = 0;
   ch[idx].old_theta = ch[idx].theta[ ch[idx].valid_data_num -1 ];
   ch[idx].old_prompt_q = ch[idx].prompt_iq[ch[idx].valid_data_num -1][1];
      
   for(int l=0; l< PULLIN_CHECK_TIME; l++)
   {
    ch[idx].delta_theta[l] = 0.;
    ch[idx].pullin_prompt_mag[l] =0.;
   }
  }
 } // end of if(ch[idx].n_confirm == 7)
}

void  GPS_Nav::ch_pullin(int idx)
{
 double cd_f, ca_f, ca_theta,ca_t, theta_err, tmp_theta;
 bool sign_changed, mag_bigenough;
 int t_idx;
 assert( idx < CH_NUM && idx >=0);

 // init total_theta to 0
 ch[idx].total_theta = 0;

 // calculate sum_theta & sum_prompt_mag for data 1 to valid_data_num
 for( int l = 0; l< ch[idx].valid_data_num; l++)
 {
  t_idx = ch[idx].pullin_time_idx;

  theta_err = fabs(ch[idx].theta[l] - sign(ch[idx].theta[l])*PI/2);

  ch[idx].sum_delta_theta +=(theta_err - ch[idx].delta_theta[t_idx] );
  ch[idx].delta_theta[ t_idx ] = theta_err;

  ch[idx].sum_prompt_mag += ( ch[idx].prompt_mag[l] -
                            ch[idx].pullin_prompt_mag[t_idx]);
  ch[idx].pullin_prompt_mag[t_idx] = ch[idx].prompt_mag[l];

  ch[idx].pullin_time_idx = (t_idx + 1)%PULLIN_CHECK_TIME;

  // calculate total_theta from 1 to valid_data_num
  if( l == 0)
   ca_t =  ch[idx].theta[l] - ch[idx].old_theta;
  else
   ca_t = ch[idx].theta[l] - ch[idx].theta[l-1];

  if( fabs(ca_t) > PI )
   ca_theta = ca_t > 0? ca_t - 2*PI : ca_t + 2*PI;
  else
   ca_theta = ca_t;
  ch[idx].total_theta += ca_theta;

  // check ms_count and ms_set
  ch[idx].ms_count = (++ch[idx].ms_count) % 20;

  sign_changed = mag_bigenough = false;
  if( l == 0)
  {
   sign_changed = (sign(ch[idx].prompt_iq[l][1]) == -sign(ch[idx].old_prompt_q));
   mag_bigenough= (fabs( ch[idx].prompt_iq[l][1]) > 800)
                &&(fabs( ch[idx].old_prompt_q) > 800);
   tmp_theta = ch[idx].old_theta;
  }
  else
  {
   sign_changed = (sign(ch[idx].prompt_iq[l][1]) == -sign(ch[idx].prompt_iq[l-1][1]));
   mag_bigenough= (fabs( ch[idx].prompt_iq[l][1]) > 800)
                &&(fabs( ch[idx].prompt_iq[l-1][1]) > 800);
   tmp_theta = ch[idx].theta[l-1];
  }

  if( sign_changed && mag_bigenough
     && fabs( fabs(ch[idx].theta[l]) - PI/2 ) < PI/8
     && fabs( fabs(tmp_theta) - PI/2 ) < PI/8 )
  {
   ch[idx].ms_count =0;
   ch[idx].ms_set = 1;
  }
 }   // end of for( int l = 0;

 ch[idx].old_prompt_q = ch[idx].prompt_iq[ch[idx].valid_data_num -1][1];

 // code pullin loop
 cd_f = (ch[idx].cd_phase1 - ch[idx].cd_phase0)* pull_code_d;
 ch[idx].code_freq = (int)(ch[idx].cd_phase1 + cd_f) + ch[idx].code_freq;
 ch_code( idx,ch[idx].code_freq);
 ch[idx].cd_phase0 = ch[idx].cd_phase1;

 // carrier pullin loop
 ca_f = ch[idx].total_theta * pull_carrier_d;
 ch[idx].carrier_freq = (int)(pull_carrier_k*ch[idx].theta[ ch[idx].valid_data_num -1 ] + ca_f) + ch[idx].carrier_freq;
 ch_carrier(idx,ch[idx].carrier_freq);
 ch[idx].old_theta = ch[idx].theta[ ch[idx].valid_data_num - 1 ];

 ch[idx].pull_time_count += ch[idx].valid_data_num;


 // after 2000ms check if we can jump to TRACK
 if( ch[idx].pull_time_count > 500)
 {
  if( ch[idx].sum_prompt_mag > 1e5 &&
      ch[idx].sum_delta_theta < 80 &&
      ch[idx].ms_set )
  {
   int valid_datanum = ch[idx].valid_data_num;
   double i_sum, q_sum;
   i_sum = ch[idx].prompt_iq[valid_datanum-1][0];
   q_sum = ch[idx].prompt_iq[valid_datanum-1][1];

   ch[idx].state = TRACKING;
   ch[idx].tlm_tst = 0x0;
   ch[idx].how_tst = 0x0;
   ch[idx].subframe_i = -1;
   ch[idx].TIC_mscount = 0;

   ch[idx].q_prompt_20 = 0.;
   ch[idx].i_prompt_20 = 0.;
   ch[idx].q_track_20 = 0.;
   ch[idx].i_track_20 = 0.;

   ch[idx].sn_mag_count = 0;
   ch[idx].sn_mag_avg =0.;
   ch[idx].SigNoise = 0.0;
   ch[idx].pseudorange = 0.0;
   ch[idx].ele = -1.;
   ch[idx].azi = 0.;
   ch[idx].tr_time_HOW = -1.;

   if( ch[idx].prompt_mag[valid_datanum-1] != 0)
    ch[idx].ca_old_phase = i_sum*sign(q_sum)/ch[idx].prompt_mag[valid_datanum-1];
   ch[idx].trk_cd_freq = ch[idx].cd_phase1;
  }
 }
 // after 10000ms if we are still in PULLIN, jump back to ACQUSITION
 if( ch[idx].pull_time_count > 10000)
 {
   ch[idx].state = ACQUISITION;
   ch[idx].code =0;
   ch[idx].n_freq -=1;  // slew back 200 Hz
   ch[idx].carrier_freq = carrier_ref+(int)(4698*ch[idx].n_freq);
 }

}
void  GPS_Nav::ch_track(int idx)
{
 double trk_ca_f,total_p, tr_time_frac;
 bool  tic_happen, tic_happen_how_rdy;
 int t_tr=0;
 int valid_datanum = ch[idx].valid_data_num;
 assert( idx < CH_NUM && idx >=0);

 ch[idx].tic_time = 0.;
 tic_happen = false;
 tic_happen_how_rdy = false;

 if( ch[idx].tr_time_HOW > 0. )  // this channel has got HOW
 {
                                // then it's time to calculate tr_time
  for( int j=0; j<4; j++)
  {
   bool new_data = (frm_4ms_6ch.all_ch[j].accum_status_a) & (ch_ready_mask[idx+CH_OFFSET]);
   if( new_data ) // new accum data for channel idx
    t_tr++;
   if( (frm_4ms_6ch.all_ch[j].accum_status_b) & 0x2000 ) // TIC happened here
    {
     tic_happen = true;
     if( new_data )
      ch[idx].tic_time = t_tr;
     else
      ch[idx].tic_time = t_tr + 0.5;
    }

  } // end of for( int j

  if( tic_happen )
  {
     if( ch[idx].tic_time == 0.5 )
     {
      int  ms_diff, tic_ms_mod20, compensate_ms;


      ms_diff = (frm_4ms_6ch.ch_cntl[idx].epoch) & 0xff;
      ms_diff -= ch[idx].epoch_ref;
      
      if(ms_diff < 0) ms_diff += 20;
      tic_ms_mod20 = ch[idx].TIC_mscount %20;
      if( fabs( ms_diff - tic_ms_mod20) < 10)
       compensate_ms = ms_diff - tic_ms_mod20;
      else
      {
       if( ms_diff > tic_ms_mod20)
        compensate_ms = -(tic_ms_mod20 - ms_diff + 20);
       else
        compensate_ms = (ms_diff - tic_ms_mod20 + 20);
      }

      ch[idx].tr_time = ch[idx].tr_time_HOW +(ch[idx].TIC_mscount+compensate_ms)*0.001+
                        (frm_4ms_6ch.ch_cntl[idx].code_phase)/2.046e6+
                        (frm_4ms_6ch.ch_cntl[idx].cd_dco_phase)/2.046e6/1024.;
                        
#if GPS_NAV_RAW_DATA == 1
     ch_rawdata_file[idx].precision(16);
     ch_rawdata_file[idx] <<dec<< ch[idx].TIC_mscount%20 << " "
                          << ch[idx].tr_time << " "
                          << ms_diff<< " "
                          << compensate_ms << " "
                          << int(ch[idx].epoch_ref)
                          << endl;
#endif
     }
     else
     {
      tr_time_frac = ch[idx].tic_time - 1.;  // possible value: 1,1.5,2,2.5,3,3.5,4
      tic_happen_how_rdy = true;
     }
    tic_happen = false;
  } // end of if( tic_happen )
 }



 total_p=0.;


 for(int j=0; j< valid_datanum; j++)
 {
  ch[idx].q_prompt_20 += ch[idx].prompt_iq[j][1];
  ch[idx].i_prompt_20 += ch[idx].prompt_iq[j][0];
  ch[idx].q_track_20 += ch[idx].track_iq[j][1];
  ch[idx].i_track_20 += ch[idx].track_iq[j][0];
  ch[idx].ms_count = (++ch[idx].ms_count)%20;
  ch[idx].TIC_mscount++;

  if( tic_happen_how_rdy )
  {
   if( floor(tr_time_frac) == j )
   {
    int  ms_diff, tic_ms_mod20, compensate_ms;

    ms_diff = (frm_4ms_6ch.ch_cntl[idx].epoch) & 0xff;
    ms_diff -= ch[idx].epoch_ref;  // minus epoch_check value
    if(ms_diff < 0) ms_diff += 20; // get the value of [0 -- 19]
    tic_ms_mod20 = ch[idx].TIC_mscount %20;
    if( fabs( ms_diff - tic_ms_mod20) < 10)
     compensate_ms = ms_diff - tic_ms_mod20;
    else  // ms_diff and tic_ms_mod20 should be very close,
     {
      if( ms_diff > tic_ms_mod20)
       compensate_ms = -(tic_ms_mod20 - ms_diff + 20);
      else
       compensate_ms = (ms_diff - tic_ms_mod20 + 20);
     }


    ch[idx].tr_time = ch[idx].tr_time_HOW+(ch[idx].TIC_mscount+compensate_ms)*0.001+
                     (frm_4ms_6ch.ch_cntl[idx].code_phase)/2.046e6 +
                     (frm_4ms_6ch.ch_cntl[idx].cd_dco_phase)/2.046e6L/1024.;
                     
#if GPS_NAV_RAW_DATA == 1
    ch_rawdata_file[idx].precision(16);
    ch_rawdata_file[idx] <<dec<< (ch[idx].TIC_mscount%20)<< " "
                         <<ch[idx].tr_time << " "
                         << ms_diff << " "
                         << compensate_ms << " "
                         << int(ch[idx].epoch_ref)
                         << endl;
#endif

    tic_happen_how_rdy = false;
   }
  }


  if( j>=1 ) // check and reset ms_count
  {
   if( ch[idx].prompt_iq[j][1]*ch[idx].prompt_iq[j-1][1] < -250000.
             && fabs( ch[idx].prompt_iq[j][1]) > 500.
             && ch[idx].prompt_iq[j-1][1] > 500.)
   {
    if( ch[idx].ms_count != 0)
     ch[idx].ms_count = 0;

    if( epoch_chk )
    {
     ch[idx].epoch_ref = ch[idx].epochchk[j] & 0xff;
    }
   }
  }
  if( ch[idx].ms_count == 19) // it's time to set code frequency
  {
   double prompt_mag, track_mag,cd_df;
   prompt_mag = sqrt(pow((float)(ch[idx].i_prompt_20),2)+pow((float)(ch[idx].q_prompt_20),2));
   track_mag = sqrt(pow((float)(ch[idx].i_track_20),2)+pow((float)(ch[idx].q_track_20),2));
   cd_df = (prompt_mag - track_mag)*200./(prompt_mag+track_mag)*trk_code_k;
   ch[idx].trk_cd_freq += cd_df;
   ch[idx].code_freq = (int)((ch[idx].trk_cd_freq/trk_code_d + cd_df)/256. -
                        (carrier_ref-ch[idx].carrier_freq)/cc_scale)
                        + code_ref;
   ch_code( idx,ch[idx].code_freq);

   ch[idx].sn_mag_avg += (prompt_mag + track_mag)/2.;

   // save TLM and HOW and try to find sync
   ch[idx].msg_count = (++ ch[idx].msg_count)%302;  // save msg to nav_msg
   ch[idx].nav_msg[ ch[idx].msg_count ] = (ch[idx].tlm_tst & 0x20000000);

   if( ch[idx].how_tst & 0x20000000 )  // check 30th bit of how_tst
    ch[idx].tlm_tst = ch[idx].tlm_tst*2 + 1;
   else
    ch[idx].tlm_tst = ch[idx].tlm_tst*2;
   // q_prompt_20 is the data bit , pipe into how

   ch[idx].how_tst = ch[idx].how_tst *2 + bit_sign(ch[idx].q_prompt_20);

   // try to find prem and calc parity
   search_pream( idx );

   if(++ch[idx].sn_mag_count == 15) // it's time to calculate SigNoise(every 15  data bits )
   {
    ch[idx].SigNoise = ch[idx].sn_mag_avg;
    if( ch[idx].SigNoise < 55000. ) // S/N less than 25dB
    {
     ch[idx].state = ACQUISITION;
     ch[idx].code =0;
     ch[idx].n_freq = (int)((ch[idx].carrier_freq - carrier_ref)/4698) -2 ;  // slew back 400 Hz
     ch[idx].carrier_freq = carrier_ref+(int)(4698*ch[idx].n_freq);
     ch[idx].subframe_i = -1;
     ch[idx].tlm_tst = 0x0;
     ch[idx].how_tst = 0x0;
     ch[idx].SigNoise = 0.0;
     ch[idx].pseudorange = 0.0;
     ch[idx].ele = 0.;
     ch[idx].azi = 0.;

    }

    ch[idx].sn_mag_count = 0;
    ch[idx].sn_mag_avg = 0.;
   }
   ch[idx].q_prompt_20 = 0.;
   ch[idx].i_prompt_20 = 0.;
   ch[idx].q_track_20 = 0.;
   ch[idx].i_track_20 = 0.;
  }

  total_p  += ch[idx].ca_phase[j];
 }

 // carrier tracking loop
 trk_ca_f = (ch[idx].ca_phase[valid_datanum-1] - ch[idx].ca_old_phase) * trk_carrier_d;
 ch[idx].carrier_freq = (int)(trk_carrier_k*total_p + trk_ca_f) + ch[idx].carrier_freq;
 ch_carrier(idx,ch[idx].carrier_freq);
 ch[idx].ca_old_phase = ch[idx].ca_phase[valid_datanum - 1];

}


// check if we get the sync
void GPS_Nav::search_pream( int idx)
{
 unsigned int TOWs,HOW,TLM,TLMs,parity0,parity1,frm;
 char d29,d30;
 d29 = (ch[idx].tlm_tst&0x80000000)>>31;
 d30 = (ch[idx].tlm_tst&0x40000000)>>30;

//#if GPS_NAV_RAW_DATA == 2
// ch_rawdata_file[idx]<< hex << ch[idx].tlm_tst << " "
//                            << ch[idx].how_tst << " "
//                            << endl;
//#endif


  if (ch[idx].tlm_tst & 0x40000000)  //
  {                                  //
   TLM = ch[idx].tlm_tst^0x3fffffc0; // then data bit reverse
  }
  else
  {
   TLM = ch[idx].tlm_tst;            // Otherwise, data bit unchanged
  }

  if (ch[idx].how_tst & 0x40000000)  //
  {                                  //  then data bit reverse
   HOW = ch[idx].how_tst^0x3fffffc0;
  }
  else
  {
                                    // Otherwise, data bit unchanged
   HOW = ch[idx].how_tst;
  }
                                           // then check parity

  if (((NAG_PREAM^TLM)& 0x3fc00000) == 0x0) // preamble pattern found?
  {
            parity0=(xors(TLM & pb1)*32)+(xors(TLM & pb2)*16)+
			  (xors(TLM & pb3)*8)+(xors(TLM & pb4)*4)+
			  (xors(TLM & pb5)*2)+(xors(TLM & pb6));


	  if (parity0 ==(TLM & 0x3f))  // is parity of TLM ok?
	  {
		  parity1=(xors(HOW & pb1)*32)+(xors(HOW & pb2)*16)+
			  (xors(HOW & pb3)*8)+(xors(HOW & pb4)*4)+
			  (xors(HOW & pb5)*2)+(xors(HOW & pb6));



		  if (parity1==(HOW & 0x3f))  // is parity of HOW ok?
		 {
                    // Get the correct sync , TLM & HOW
                   frm = int((HOW & 0x700)>>8);
                   if( frm > 0 && frm < 6)  // valid sub frame #?
                   {
#if GPS_NAV_RAW_DATA == 2
                  ch_rawdata_file[idx]<<dec
                                      << TIC_count << "* "
                                      << ch[idx].TIC_mscount << "* "
                                      << ch[idx].subframe_i << " "
                                      << endl;
#endif

                    if( ch[idx].TIC_mscount % 6000 == 0) // this is a correct pream ?
                    {
                     ch[idx].subframe_i = frm;
                     ch[idx].TOW =(HOW & 0x3fffffffL)>>13;
                     ch[idx].TLM =(TLM>>8) & 0x3fff;
                     ch[idx].tr_time_HOW = (ch[idx].TOW-1)*6. + 1.2;

                     if( !newfrmflag ) // is it the first frame?
                     {
                      newfrmcount = 0;       // reset newfrmcount to zero
                      newfrmflag = true;     // set newfrmflag to true
                     }
                     // then copy msg

                     allch_navmsg[idx].newdata = true;
                     allch_navmsg[idx].prn = ch[idx].prn;
                     allch_navmsg[idx].index =  ch[idx].msg_count;
                     memcpy(&allch_navmsg[idx].msg[0], &ch[idx].nav_msg[0], 302*sizeof( unsigned int ));
                   }
                   ch[idx].TIC_mscount = 0;  //
                  }
                 }
          }
  }
}

// allocate prn to channel 1-6
// pn is the pointer pointed to prn numbers
void  GPS_Nav::SetPrn(int*  pn)
{
 for(int i =0; i<CH_NUM; i++)
  ch[i].prn = pn[i];
}

void GPS_Nav::init_gp2021(void)
{
 int prn_idx;
 // first package to initialize ,
 // setup io_config, system_setup, enable 8 channels
 // then set PRN code for these 8 channels
 clear_all_reg();           // every time to write one packet, need to clear all values
 reset_cntl(0x0);           // 0x7f
 io_config(0x0301);         // 0xf0
 test_control(0x0);         // 0x7c
 system_setup(0x0);         // 0x7e
 reset_cntl(ENABLE_6CH);    // 0x7f
 test_data(0xccdd);         // for debug purpose

 IoHandle->GetCntlData(0,(unsigned char*) &gp2021_wr_data); // buffer0
 IoHandle->rd_wr_gp2021_all(0,false,NULL); // send to GP2021, buffer0, no returen

 Sleep(50);
 // second package to initialize
 // setup the carrier and code DCO frequency

 clear_all_reg();          // every time to write one packet, need to clear all values
 for(char i=0; i<CH_NUM; i++){
         ch_code( i,code_ref);
         ch_carrier(i,carrier_ref+(long)(4698*ch[i].n_freq));
         ch[i].carrier_freq = carrier_ref+(int)(4698*ch[i].n_freq);
         prn_idx = ch[i].prn;
         ch_cntl(i, prn_code[prn_idx] | 0xa000);
}

 IoHandle->GetCntlData(0,(unsigned char*) &gp2021_wr_data); // buffer0
 IoHandle->rd_wr_gp2021_all(0,false,NULL); // send to GP2021, buffer0, no returen

 Sleep(50);

 clear_all_reg();
}
/* end of GP2021  related function **/
/***************************************************************************/


//---------------------------------------------------------------------------
/*    Routine Description:                                                 */
/*  Init proc for class;                                                   */
/*  Return non-zero if something wrong, otherwise return 0                 */
/*  Error code is defined in gbl_var.h                                     */
/***************************************************************************/
int GPS_Nav::InitProc()
{
#if GPS_NAV_RAW_DATA > 0
  char tmpbuf[80];
  // open raw data file for channels
  for(int i=0; i<CH_NUM; i++)
  {
   sprintf(tmpbuf, ".\\log\\ch%d_raw.dat", i);
   ch_rawdata_file[i].open(tmpbuf, ios::out);
  }

  // now open accum_int and accum_mis file object
  sprintf(tmpbuf, ".\\log\\ch_accum_int.dat");
  sprintf(tmpbuf, ".\\log\\ch_accum_mis.dat");
  ch_accum_mis_file.open(tmpbuf,ios::out);

#endif

  for(int i=0; i<CH_NUM; i++){
   ch[i].code = 0;
   ch[i].n_confirm =0;
   ch[i].state = ACQUISITION;
   ch[i].n_freq = SEARCH_MIN_F;
   ch[i].del_freq = 1;
   ch[i].rawdata.ch_num = i;
   ch[i].rawdata.prn_num = ch[i].prn;
   ch[i].code_freq = code_ref;
   ch[i].cd_phase0=ch[i].cd_phase1= 0.0;
   for(int j=0; j<4; j++)         
    ch[i].theta[j] = 0.0;
   data_count[i] = 0;

   ch[i].msg_count = 0;
   memset( ch[i].nav_msg, 0, sizeof( ch[i].nav_msg ));
  }

#ifdef GPS_NAV_DBG
  debugfile.open(".\\log\\gps_nav.log",ios::out);
  if(!debugfile)
  return(GPS_NAV_FILEOPENERR);
#endif

  return (GPS_NAV_SUCCESS);
}
//---------------------------------------------------------------------------
/*    Routine Description:                                                 */
/*  print debug info for class;                                            */
/***************************************************************************/

void GPS_Nav::PrintDbgInfo(char* info)
{

#ifdef GPS_NAV_DBG
   if(debugfile)
    debugfile << info << endl;
#endif

}
//---------------------------------------------------------------------------
/*    Routine Description:                                                 */
/*  get data from sampling thread and save to local   data buffer          */
/***************************************************************************/
void GPS_Nav::Get_oneframedata()
{
 count ++;
 IoHandle->CopyData((unsigned char*)&frm_4ms_6ch);
}
//---------------------------------------------------------------------------
/*    Routine Description:                                                 */
/*  process data for current frame                                         */
/***************************************************************************/
int GPS_Nav::ProcessData()
{
 char buf[80];
 int  tic;


 tic = 0;
 count1 = (count1+1)%16; // for debugging

 if(frm_4ms_6ch.timestamp[0] != frm_4ms_6ch.timestamp[3])
 {
  sprintf(buf, "%4x * %4x * %4x *%4x @ %6d ",
            (int)frm_4ms_6ch.timestamp[0],
            (int)frm_4ms_6ch.timestamp[1],
            (int)frm_4ms_6ch.timestamp[2],
            (int)frm_4ms_6ch.timestamp[3],
             count);
  PrintDbgInfo(buf);
 }

#if GPS_NAV_RAW_DATA > 0
/* here write into accum_mis_file ACCUM_STATUS_B every frame
 ch_accum_mis_file.width(4);
 ch_accum_mis_file << hex
                   << frm_4ms_8ch.all_ch[0].accum_status_b << " "
                   << frm_4ms_8ch.all_ch[1].accum_status_b << " "
                   << frm_4ms_8ch.all_ch[2].accum_status_b << " "
                   << frm_4ms_8ch.all_ch[3].accum_status_b << " "
                   << (int)((frm_4ms_6ch.dummy[0] & 0xf0)>>4) <<" "
                   << (int)count1<< " "
                   << (int)(frm_4ms_6ch.dummy[1]) << " "
                   << (int)(frm_4ms_6ch.dummy[2])
                   << endl;
 // here dummy[0] indicate the result of which command
 //      count1 indicate current command timestamp
 //      dummy[1] indicate the frm count of this result
 //      dummy[2] indicate the frm  subcount of this result frame
 // all these data are used for debugging
*/

/* here write into accum_mis_file ACCUM_STATUS_B only when timeout happens
*/
 for(int j=0; j<4; j++)
  if( frm_4ms_6ch.all_ch[j].accum_status_b & 0x3f )
  {
   ch_accum_mis_file.width(4);
   ch_accum_mis_file << hex
                   << frm_4ms_6ch.all_ch[0].accum_status_b << " "
                   << frm_4ms_6ch.all_ch[1].accum_status_b << " "
                   << frm_4ms_6ch.all_ch[2].accum_status_b << " "
                   << frm_4ms_6ch.all_ch[3].accum_status_b << " "
                   << endl;
   break;
  }
#endif

 clear_all_reg();

 // first get new accum data and calculate norm, phase,etc
 get_newaccumdata();
 // then process different channel according to new accum data
 for(int ch_i=0; ch_i< CH_NUM; ch_i++)
  process_channel(ch_i);


#if GPS_NAV_RAW_DATA == 3

 for(int i=0; i< CH_NUM; i++)
 {
 // ch_rawdata_file[i]<< (int)ch[i].valid_data_num << " ";
  if( ch[i].state >= PULLIN )
   for( int k=0; k < ch[i].valid_data_num; k++)
   {
    ch_rawdata_file[i] << ch[i].prompt_iq[ k ][0] << " "
                       << ch[i].prompt_iq[ k ][1] << " "
                       //<< ch[i].track_iq[ k ][0] << " "
                       //<< ch[i].track_iq[ k ][1] << " "
                       << ch[i].theta[k]<< " "
                       //<< ch[i].sum_prompt_mag << " "
                       //<< ch[i].sum_delta_theta << " "
                       //<< ch[i].total_theta << " "//ca_phase[k] << " "
                       << (ch[i].code_freq - code_ref)*0.08514949<< " "
                       << (ch[i].carrier_freq - carrier_ref)* 0.04257475 <<" "
                       << ch[i].total_theta<<" "
                       << ch[i].ms_count << " "
                       << endl;
   }
 }

#endif



 for(int j=0; j<4; j++)
 {
   if( (frm_4ms_6ch.all_ch[j].accum_status_b) & 0x2000 ) // TIC happened here
   {

    allchtrtime.ch_count = 0;

    if( (++newfrmcount % NAV_UPDATE_INT) == 2)  // every 2sec to trigger gpsFunc thrd
    {
     if( newfrmflag )  // if new frame data is get, no matter how many channels
     {
      newmsg_rdy = true;     // GpsFunc class should process msg & calculate PVT
      newfrmflag = false;    // reset newfrmflag to not valid
     }
     else
      newmsg_rdy = false;    // GpsFunc class only need to calculate PVT

     int ch_cnt ;

     for( int i=0; i<CH_NUM; i++)
     {
      if( ch[i].state == TRACKING && ch[i].SigNoise > 111500.
          && ch[i].tr_time_HOW > 0 && eph_valid[ ch[i].prn ] )
      {
       ch_cnt = allchtrtime.ch_count++;
       allchtrtime.ch_transtime[ch_cnt].prn = ch[i].prn;
       allchtrtime.ch_transtime[ch_cnt].tr_time =ch[i].tr_time;
       allchtrtime.ch_transtime[ch_cnt].dopp_f = (ch[i].carrier_freq - carrier_ref)* 0.04257475;
      }
     }
     SetEvent(gpsEvent);    // inform GpsFunc class to process msg
    }// end of if( (++newfrmcount % NAV_UPDATE_INT) == 2)


    for(int i=0; i<CH_NUM; i++)
    {
     ch[i].rawdata.code_slew = frm_4ms_6ch.ch_cntl[i].code_slew;
     ch[i].rawdata.code_phase = frm_4ms_6ch.ch_cntl[i].code_phase;
     ch[i].rawdata.carrier_cycle_high = frm_4ms_6ch.ch_cntl[i].ca_cycle_high;
     ch[i].rawdata.carrier_cycle_low = frm_4ms_6ch.ch_cntl[i].ca_cycle_low;
     ch[i].rawdata.carrier_dco_phase = frm_4ms_6ch.ch_cntl[i].ca_dco_phase;
     ch[i].rawdata.code_dco_phase = frm_4ms_6ch.ch_cntl[i].cd_dco_phase;
     ch[i].rawdata.epoch =  frm_4ms_6ch.ch_cntl[i].epoch;
     ch[i].rawdata.i_track = frm_4ms_6ch.all_ch[j].ch_accm[i].I_track;
     ch[i].rawdata.q_track = frm_4ms_6ch.all_ch[j].ch_accm[i].Q_track;
     ch[i].rawdata.i_prompt= frm_4ms_6ch.all_ch[j].ch_accm[i].I_prompt;
     ch[i].rawdata.q_prompt= frm_4ms_6ch.all_ch[j].ch_accm[i].Q_prompt;
     ch[i].rawdata.freq_off = ch[i].n_freq*200;

     tic = 1;
    }
    TIC_count++;

    EnterCriticalSection(&GpsCriticalSection);
    for( int i=0; i< allchtrtime.ch_count; i++)
    {

     if( allch_navresult[i].valid )
     {
      for( int k=0; k<CH_NUM; k++)
       if( ch[k].prn == allch_navresult[i].prn)
       {
        ch[k].pseudorange = allch_navresult[i].pseudorange;
        ch[k].ele = allch_navresult[i].ele;
        ch[k].azi = allch_navresult[i].azi;
       }
      allch_navresult[i].valid = false;
     } //end of  if( allch_navresult[i].valid )
    } // end of  for( int i=0; i< allchtr
    LeaveCriticalSection(&GpsCriticalSection);
   }// end of if( (frm_4ms_6ch.all_ch[j].accum_status_b) & 0x2000 )
 }

 // copy control data to interface 's data buffer
 gp2021_wr_data.flag = count1 << 4;
 IoHandle->GetCntlData(IoHandle->get_free_pp_flag(),
                      (unsigned char*) &gp2021_wr_data);
 return (tic);
}
//---------------------------------------------------------------------------
/*    Routine Description:                                                 */
/*    thread handling funtcions                                          */
/***************************************************************************/
void GPS_Nav::CreateNavThrd()
{
 clear_all_reg();
 IoHandle->CreatSampleThrd();
 navthrd = new NavThread(true, this);
 thrdtermed = false;
 navthrd->Resume();
}
 // terminate sampling thread
void GPS_Nav::TerminateThrd()
{
 if(navthrd)
  navthrd->Terminate();
}
void GPS_Nav::ResumeThrd()
{
 if(navthrd)
 navthrd->Resume();
}
void GPS_Nav::SuspendThrd()
{
 if(navthrd)
 navthrd->Suspend();
}
 // return the terminate status of sampling thread
bool GPS_Nav::NavThrdIsTermed()
{
 return (thrdtermed);
}
 // function to set the terminate status of sampling thread
void GPS_Nav::SetThrdStatus(bool stat)
{
 thrdtermed = stat;
}
// to copy raw data for display use
void GPS_Nav::get_rawdata( PT_allch_rawdata_disp rawdata_pt )
{
 for( int i=0; i<CH_NUM; i++)
  rawdata_pt->allch_rawdata[i] = ch[i].rawdata;
}

void GPS_Nav::get_trkloop(PT_channel_disp res_pt)
{
 for(int i=0; i<CH_NUM; i++)
 {
  res_pt[i].prn_num = ch[i].prn;
  res_pt[i].state_info = ch[i].state;
  res_pt[i].freq_off = (ch[i].carrier_freq - carrier_ref)* 0.04257475;
  res_pt[i].SigNoise = ch[i].SigNoise;
  res_pt[i].pseudorange = ch[i].pseudorange;
  res_pt[i].frm_idx = ch[i].subframe_i;
  res_pt[i].tow = ch[i].TOW;
  res_pt[i].tlm = ch[i].TLM;
  res_pt[i].ele = ch[i].ele;
  res_pt[i].azi = ch[i].azi;
 }
}





